#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import argparse
import getpass
import logging
import pprint
import sys
import uuid
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, TypeVar, cast

import dbus
import NetworkManager as nm

device_list: Optional[List[nm.Device]] = None
settings_cache: Dict[str, Any] = {}

pp = pprint.PrettyPrinter(indent=4)

log = logging.getLogger("comitup")


def initialize() -> None:
    nm.Settings.ReloadConnections()


A = TypeVar("A")
R = TypeVar("R")


# TODO - use ParamSpec["A"] once 3.10 become ubiquitous
def none_on_exception(
    *exceptions,
) -> Callable[[Callable[[A], R]], Callable[[A], Optional[R]]]:
    def _none_on_exception(fp: Callable[..., R]) -> Callable[..., Optional[R]]:
        @wraps(fp)
        def wrapper(*args, **kwargs) -> Optional[R]:
            try:
                return fp(*args, **kwargs)
            except exceptions:
                log.debug("Got an exception, returning None, %s", fp.__name__)
                return None

        return wrapper

    return _none_on_exception


def get_devices() -> Optional[List[nm.Device]]:
    global device_list

    if not device_list:
        try:
            device_list = nm.NetworkManager.GetDevices()
        except TypeError:
            # NetworkManager is gone for some reason. Bail big time.
            sys.exit(1)

    if not device_list:
        log.error("No devices found in nm.get_devices()")

    return device_list


def device_name(device: nm.Device) -> str:
    return device.Interface


def get_wifi_devices() -> List[nm.Wireless]:
    devices: Optional[List[nm.Device]] = get_devices()

    if devices is not None:
        devices = [x for x in devices if x.DeviceType == 2]
        return [cast(nm.Wireless, x) for x in devices]
    else:
        log.error("No WiFi devices found in nm.get_wifi_devices()")
        return []


def get_phys_dev_names() -> List[str]:

    devices = get_devices()

    if devices is not None:
        return [device_name(x) for x in devices if x.DeviceType in (1, 2)]
    else:
        log.error("No devices found in nm.get_phys_dev_names")
        return []


@none_on_exception(IndexError)
def get_wifi_device(index: int = 0) -> Optional[nm.Device]:
    return get_wifi_devices()[index]


def get_device_path(device: nm.Device) -> dbus.ObjectPath:
    log.debug("Getting specific device for path")
    return device.SpecificDevice().object_path


def disconnect(device: nm.Device) -> None:
    try:
        log.debug("Calling disconnect")
        device.Disconnect()
    except:  # noqa
        log.debug("Error received in disconnect")


def get_connection_settings(connection: nm.Connection) -> Dict[str, Any]:
    global settings_cache

    if connection.uuid not in settings_cache:
        log.debug("Not in cache")
        settings_cache[connection.uuid] = connection.GetSettings()

    return settings_cache[connection.uuid]


def get_device_settings(device: nm.Device) -> Dict[str, Any]:
    try:
        connection = device.ActiveConnection
    except nm.ObjectVanished:
        sys.exit(1)

    return get_connection_settings(connection.Connection)


@none_on_exception(AttributeError)
def get_active_ssid(device: nm.Device) -> Optional[str]:
    return get_device_settings(device)["802-11-wireless"]["ssid"]


@none_on_exception(AttributeError, IndexError)
def get_active_ip(device: nm.Device) -> Optional[str]:
    addr: str = device.Ip4Address
    if addr == "0.0.0.0":
        addr = device.Ip4Config.Addresses[0][0]

    return addr


@none_on_exception(AttributeError, IndexError)
def get_active_ip6(device: nm.Device) -> Optional[str]:
    addr6 = device.Ip6Config.Addresses[0][0]

    return addr6


def get_all_connections() -> List[nm.Connection]:
    return [x for x in nm.Settings.ListConnections()]


def get_all_wifi_connection_ssids():
    for conn in get_all_connections():
        ssid = get_ssid_from_connection(conn)
        if ssid:
            yield ssid


@none_on_exception(AttributeError, KeyError)
def get_ssid_from_connection(connection: nm.Connection) -> Optional[str]:
    settings = get_connection_settings(connection)

    return settings["802-11-wireless"]["ssid"]


def get_connection_by_ssid(name: str) -> Optional[nm.Connection]:
    for connection in get_all_connections():
        ssid = get_ssid_from_connection(connection)
        if name == ssid:
            return connection

    return None


def del_connection_by_ssid(name: str) -> None:
    for connection in get_all_connections():
        ssid = get_ssid_from_connection(connection)
        if name == ssid:
            connection.Delete()


def activate_connection_by_ssid(ssid: str, device: nm.Device, path: str = "/"):
    connection = get_connection_by_ssid(ssid)

    if connection:
        log.debug("    {}".format(get_ssid_from_connection(connection)))
        log.debug("    {}".format(device_name(device)))

        opath = nm.NetworkManager.ActivateConnection(connection, device, path)

        log.debug("ActivateConnection returned {}".format(opath.object_path))
    else:
        log.error("Connection for {} not found".format(ssid))


@none_on_exception(AttributeError)
def get_access_points(device: nm.Device) -> Optional[List[nm.AccessPoint]]:
    return device.SpecificDevice().GetAllAccessPoints()


def get_candidate_connections(device: nm.Device) -> List[str]:
    candidates = []

    for conn in get_all_connections():
        log.debug("Getting settings 2")
        settings = get_connection_settings(conn)
        ssid = get_ssid_from_connection(conn)

        try:
            if (
                ssid
                and settings["connection"]["type"] == "802-11-wireless"
                and settings["802-11-wireless"]["mode"] == "infrastructure"
            ):

                candidates.append(ssid)
        except KeyError:
            log.debug("Unexpected connection format for %s" % ssid)

    log.debug("candidates: %s" % candidates)

    return candidates


def make_hotspot(name="comitup", device=None, password="", hash="0000"):
    comitup_uuid = str(uuid.uuid4())
    settings = {
        "connection": {
            "type": "802-11-wireless",
            "uuid": comitup_uuid,
            "id": "{0}-{1}".format(name, hash),
            "autoconnect": False,
        },
        "802-11-wireless": {
            "mode": "ap",
            "ssid": name,
            "band": "bg",
        },
        "ipv4": {
            "method": "manual",
            "address-data": [
                {
                    "address": "10.41.0.1",
                    "prefix": 24,
                }
            ],
        },
        "ipv6": {
            "method": "ignore",
        },
    }

    if device:
        settings["connection"]["interface-name"] = device_name(device)

    if password and len(password) >= 8:
        settings["802-11-wireless-security"] = {}
        settings["802-11-wireless-security"]["key-mgmt"] = "wpa-psk"
        settings["802-11-wireless-security"]["psk"] = password

    nm.Settings.AddConnection(settings)


def make_connection_for(
    ssid: str,
    password: str = None,
    interface: Optional[str] = None,
    link_local: bool = True,
) -> None:
    settings = dbus.Dictionary(
        {
            "connection": dbus.Dictionary(
                {
                    "id": ssid,
                    "type": "802-11-wireless",
                    "uuid": str(uuid.uuid4()),
                }
            ),
            "802-11-wireless": dbus.Dictionary(
                {
                    "ssid": dbus.ByteArray(ssid.encode()),
                    "mode": "infrastructure",
                }
            ),
            "ipv4": dbus.Dictionary(
                {
                    # assume DHCP
                    "method": "auto",
                }
            ),
            "ipv6": dbus.Dictionary(
                {
                    "method": "link-local",
                }
            ),
        }
    )

    if not link_local:
        settings["ipv6"]["method"] = "auto"

    if interface:
        settings["connection"]["interface-name"] = interface

    # assume privacy = WPA(2) psk
    if password:
        settings["802-11-wireless"]["security"] = "802-11-wireless-security"
        settings["802-11-wireless-security"] = dbus.Dictionary(
            {
                "auth-alg": "open",
                "key-mgmt": "wpa-psk",
                "psk": password,
            }
        )

    nm.Settings.AddConnection(settings)


#
# CLI Interface
#


def do_listaccess(arg):
    """List all accessible access points"""
    rows = []
    for point in get_access_points(get_wifi_devices()[-1]):
        row = (
            point.Ssid,
            point.HwAddress,
            point.Flags,
            point.WpaFlags,
            point.RsnFlags,
            point.Strength,
            point.Frequency,
        )
        rows.append(row)

    bypwr = sorted(rows, key=lambda x: -x[5])

    hdrs = ("SSID", "MAC", "Private", "WPA", "RSN", "Power", "Frequency")

    try:
        import tabulate

        print(tabulate.tabulate(bypwr, headers=hdrs))
    except:  # noqa
        for entry in bypwr:
            print(entry)


def do_listconnections(arg):
    """List all defined connections"""
    for connection in get_all_connections():
        ssid = get_ssid_from_connection(connection)

        if ssid:
            print(ssid)


def do_setconnection(ssid):
    """Connect to a connection"""
    activate_connection_by_ssid(ssid, get_wifi_device())


def do_getconnection(dummy):
    """Print the active connection ssid"""
    print(get_active_ssid(get_wifi_device()))


def do_getip(dummy):
    """Print the current IP address"""
    print(get_active_ip(get_wifi_device()))


def do_detailconnection(ssid):
    """Print details about a connection"""

    connection = get_connection_by_ssid(ssid)

    pp.pprint(connection.GetSettings())
    pp.pprint(connection.GetSecrets())


def do_delconnection(ssid):
    """Delete a connection id'd by ssid"""

    del_connection_by_ssid(ssid)


def do_makehotspot(dummy):
    """Create a hotspot connection for future use"""

    make_hotspot()


def do_listcandidates(dummy):
    """List available connections for current access points"""

    for candidate in get_candidate_connections(get_wifi_device()):
        print(candidate)


def do_makeconnection(ssid):
    """Create a connection for an access point, for future use"""

    password = getpass.getpass("password: ")

    make_connection_for(ssid, password)


def get_command(cmd):
    cmd_name = "do_" + cmd

    try:
        return globals()[cmd_name]
    except KeyError:
        return None


def parse_args():
    prog = "comitup"
    epilog = "Commands:\n"
    for x in sorted([x for x in globals().keys() if x[0:3] == "do_"]):
        epilog += "  %s - %s\n" % (x[3:], globals()[x].__doc__)

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        prog=prog,
        description="Manage NetworkManager Wifi connections",
        epilog=epilog,
    )

    parser.add_argument(
        "command",
        help="command",
    )

    parser.add_argument(
        "arg",
        nargs="?",
        help="command argument",
    )

    args = parser.parse_args()

    if get_command(args.command) is None:
        parser.error("Invalid command")

    return args


def main():
    try:
        initialize()
    except dbus.exceptions.DBusException:
        print("Must run as root")
        sys.exit(1)

    args = parse_args()
    fn = get_command(args.command)
    fn(args.arg)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import argparse
import sys
from pathlib import Path
from re import MULTILINE, search, sub
from subprocess import run
from typing import List

sys.path.append(".")
sys.path.append("..")

from collections import OrderedDict, namedtuple  # noqa
from getpass import getpass  # noqa

from comitup import client as ciu  # noqa


def do_reload(ciu_client, connection):
    pass


def do_quit(ciu_client, connection):
    sys.exit(0)


def do_delete(ciu_client, connection):
    ciu_client.ciu_delete(connection)


def do_connect(ciu_client, ssid, password):
    ciu_client.ciu_connect(ssid, password)


def do_info(ciu_client, connection):
    info = ciu_client.ciu_info(connection)
    print(
        "Host %s on comitup version %s" % (info["hostnames"], info["version"])
    )
    print("'%s' mode" % info["imode"])

    state, _ = get_state(ciu_client)
    print("{} state".format(state))


def do_locate(ciu_client, connection):
    print("Attempting to blink the Pi front green led once")
    try:
        ciu.blink()
    except PermissionError:
        print("ERROR: Run as root")


def set_conf(key, val):
    confpath = Path("/etc/comitup.conf")

    confdata = confpath.read_text()

    if search("^{}: .+$".format(key), confdata, flags=MULTILINE):
        confdata = sub(
            "^{}: .+$".format(key),
            "{}: {}".format(key, val),
            confdata,
            flags=MULTILINE,
        )
    elif search("^# {}:.+$".format(key), confdata, flags=MULTILINE):
        confdata = sub(
            "^(# {}:.+)$".format(key),
            r"\1\n{}: {}".format(key, val),
            confdata,
            flags=MULTILINE,
        )
    else:
        confdata += "{}: {}\n".format(key, val)

    confpath.write_text(confdata)


def do_name(ciu_client, name):
    with open("/etc/hostname", "w", encoding="utf-8") as fp:
        fp.write(name)
    with open("/etc/hosts", "a", encoding="utf-8") as fp:
        fp.write("127.0.0.1\t{}\n".format(name))
    run(["/usr/bin/hostname", name])
    set_conf("ap_name", name)
    run(["/usr/bin/systemctl", "restart", "comitup"])

    sys.exit(0)


def do_nuke(ciu_client, connection):
    ciu_client.ciu_nuke()


CmdState = namedtuple(
    "CmdState", "fn, desc, HOTSPOT, CONNECTING, CONNECTED, scriptable"
)

commands = OrderedDict(
    [
        ("r", CmdState(do_reload, "(r)eload", True, True, True, False)),
        ("i", CmdState(do_info, "(i)nfo", True, True, True, True)),
        (
            "l",
            CmdState(do_locate, "(l)ocate the device", True, True, True, True),
        ),
        (
            "n",
            CmdState(
                do_name,
                "re(n)ame the device (restarts the service)",
                True,
                True,
                True,
                True,
            ),
        ),
        (
            "d",
            CmdState(
                do_delete, "(d)elete connection", False, True, True, True
            ),
        ),
        (
            "<n>",
            CmdState(do_connect, "connect to <n>", True, False, False, False),
        ),
        (
            "m",
            CmdState(
                do_connect, "(m)anual connection", True, False, False, True
            ),
        ),
        (
            "x",
            CmdState(
                do_nuke,
                "(x) Factory reset (no warning)",
                True,
                True,
                True,
                True,
            ),
        ),
        ("q", CmdState(do_quit, "(q)uit", True, True, True, False)),
    ]
)


def int_value(s):
    try:
        return int(s)
    except (ValueError, TypeError):
        return None


def get_state(ciu_client):
    state, connection = ciu_client.ciu_state()
    return state, connection


def get_valid_cmds(state):
    cmds = [x for x in commands.keys() if commands[x].__getattribute__(state)]
    return cmds


def print_cmd_prompts(state, connection, points):
    print("")
    print("State: %s" % state)
    print("Connection: %s" % connection)

    if state == "HOTSPOT":
        print("Points:")
        for point in enumerate(points, start=1):
            print("    %d: %s" % (point[0], point[1]["ssid"]))

    print("Available commands:")

    for cmd in get_valid_cmds(state):
        print("    %s" % commands[cmd].desc)


def interpreter():

    ciu_client = ciu.CiuClient()

    while True:
        state, connection = get_state(ciu_client)

        points = ciu_client.ciu_points()

        print_cmd_prompts(state, connection, points)

        cmd = input("command?: ")

        index = int_value(cmd)

        if index:
            password = ""
            if points[index - 1]["security"] == "encrypted":
                password = getpass("password: ")
            do_connect(ciu_client, points[index - 1]["ssid"], password)
        elif cmd == "m":
            ssid = input("ssid?: ")
            password = getpass("password (if required)?: ")
            do_connect(ciu_client, ssid, password)
        elif cmd == "n":
            name = input("new host name?: ")
            do_name(ciu_client, name)
        else:
            try:
                print()
                commands[cmd].fn(ciu_client, connection)
            except KeyError:
                print("\nInvalid command\n")


def one_shot(cmdlist: List[str]):
    cmd = cmdlist[0]

    if cmd not in commands:
        print("Invalid command {}".format(cmd))
        return

    if not commands[cmd].scriptable:
        print("Command is not scriptable")
        return

    ciu_client = ciu.CiuClient()

    state, connection = get_state(ciu_client)

    if cmd == "m":
        print("Attempting to connect to {}".format(cmdlist[1]))
        do_connect(ciu_client, *cmdlist[1:])
    elif cmd == "n":
        do_name(ciu_client, *cmdlist[1:])
    else:
        commands[cmd].fn(ciu_client, connection)


def parse_args():
    parser = argparse.ArgumentParser(
        description="A CLI front end for the Comitup service"
    )

    parser.add_argument("command", nargs="*", help="a one-shot command")

    return parser.parse_args()


def main():
    args = parse_args()

    if args.command:
        one_shot(args.command)
    else:
        interpreter()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE


import argparse
import logging
import os
import signal
import sys
import types
from logging.handlers import TimedRotatingFileHandler
from typing import Optional

from dbus.mainloop.glib import DBusGMainLoop

DBusGMainLoop(set_as_default=True)
from gi.repository.GLib import MainLoop  # noqa

from comitup import cdns  # noqa
from comitup import config  # noqa
from comitup import iptmgr  # noqa
from comitup import nuke  # noqa
from comitup import persist  # noqa
from comitup import statemgr  # noqa
from comitup import sysd  # noqa
from comitup import webmgr  # noqa
from comitup import wificheck  # noqa

LOG_PATH: str = "/var/log/comitup.log"
log: Optional[logging.Logger] = None


def deflog(verbose: bool) -> logging.Logger:
    level = logging.INFO
    if verbose:
        level = logging.DEBUG

    log = logging.getLogger("comitup")
    log.setLevel(level)
    handler = TimedRotatingFileHandler(
        LOG_PATH,
        encoding="utf=8",
        when="W0",
        backupCount=8,
    )
    fmtr = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(fmtr)
    log.addHandler(handler)

    return log


def check_environment(log: logging.Logger) -> None:
    for service in ["systemd-resolved", "dnsmasq", "dhcpd", "dhcpcd"]:
        try:
            if sysd.sd_unit_jobs("{}.service".format(service)):
                for msg in [
                    "Warning: {} service is active.".format(service),
                    "This may interfere with comitup providing "
                    "networking services",
                ]:
                    print(msg)
                    log.warn(msg)
        except Exception:
            pass


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="")

    parser.add_argument(
        "-c",
        "--check",
        action="store_true",
        help="Check the wifi devices and exit",
    )

    parser.add_argument(
        "-i",
        "--info",
        action="store_true",
        help="Print info and exit",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="More verbose logging",
    )

    args = parser.parse_args()

    return args


def cleanup() -> None:
    # leave the network setup as it, but kill comitup-web
    webmgr.stop_service(webmgr.COMITUP_SERVICE)
    nuke.cleanup_nuke()
    if log:
        log.info("Stopping comitup")


def handle_term(signum: int, handler: types.FrameType) -> None:
    cleanup()
    sys.exit(0)


def main():
    global log

    if os.geteuid() != 0:
        exit("Comitup requires root privileges")

    args = parse_args()

    (conf, data) = config.load_data()

    log = deflog(args.verbose or conf.getboolean("verbose"))
    log.info("Starting comitup")

    if args.info:
        for key, val in statemgr.get_info(conf, data).items():
            print("{}: {}".format(key, val))
        sys.exit(0)

    if args.check:
        if wificheck.run_checks():
            sys.exit(1)
        else:
            sys.exit(0)
    else:
        wificheck.run_checks(verbose=False)

    check_environment(log)

    webmgr.init_webmgr(conf.web_service)
    iptmgr.init_iptmgr()

    statemgr.init_state_mgr(
        conf,
        data,
        [
            webmgr.state_callback,
            iptmgr.state_callback,
            cdns.state_callback,
        ],
    )

    nuke.init_nuke()

    signal.signal(signal.SIGTERM, handle_term)

    loop = MainLoop()

    try:
        loop.run()
    except Exception:
        log.error("Terminal exception encountered")
        raise
    finally:
        cleanup()


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# Copyright (c) 2017-2021 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

from typing import TYPE_CHECKING, Optional

from comitup import config, nm

if TYPE_CHECKING:
    import NetworkManager

SINGLE_MODE: str = "single"
MULTI_MODE: str = "router"

CONF_PATH: str = "/etc/comitup.conf"

ap_device: Optional["NetworkManager.Device"] = None
link_device: Optional["NetworkManager.Device"] = None


def get_conf() -> config.Config:
    (conf, _) = config.load_data()
    return conf


def dual_enabled() -> bool:
    return get_conf().getboolean("enable_appliance_mode")


def get_mode() -> str:
    if len(nm.get_wifi_devices()) > 1 and dual_enabled():
        return MULTI_MODE
    else:
        return SINGLE_MODE


def get_ap_device() -> "NetworkManager.Device":
    global ap_device

    if not ap_device:
        devs = nm.get_wifi_devices()
        spec = get_conf().primary_wifi_device

        if spec:
            for dev in devs:
                if dev.Interface == spec:
                    ap_device = dev

    if not ap_device:
        ap_device = devs[0]

    if ap_device:
        return ap_device
    else:
        raise


def get_link_device() -> "NetworkManager.Device":
    global link_device

    if not link_device:
        devs = nm.get_wifi_devices()
        link_device = get_ap_device()

        if dual_enabled():
            for dev in devs:
                if dev.Interface != link_device.Interface:
                    link_device = dev
                    return link_device

    return link_device


def get_state_device(state: str) -> "NetworkManager.Device":
    if state == "HOTSPOT":
        return get_ap_device()
    else:
        return get_link_device()



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import logging
import re
import struct
import subprocess
from multiprocessing import Process, Queue
from typing import Dict, List

log = logging.getLogger("comitup")

# NetworkManager is doing a poor job of maintaining the AP scan list when
# in AP mode. Use the 'iw' command to get the AP list.


def docmd(cmd: str) -> str:
    cmd = "timeout 5 " + cmd

    try:
        out = subprocess.check_output(cmd.split(), encoding="utf-8")
    except subprocess.CalledProcessError:
        out = ""

    return out


def devlist() -> List[str]:
    """Get a list of supported devices from 'iw'"""
    log.debug("Getting device list from 'iw'")
    out = docmd("iw dev")
    devs = [x.split()[1] for x in out.split("\n") if "Interface" in x]
    return devs


def blk2dict(blk: str) -> Dict[str, str]:
    """Convert a 'iw dev scan' block into a tagged dict"""

    lines = [x.strip().split(":") for x in blk.split("\n") if ":" in x]
    tups = [(x[0].strip(), x[1].strip()) for x in lines if len(x) > 1]

    return dict(tups)


def dbm2pct(dbm: float) -> str:
    pct = (dbm + 100.0) * 2
    pct = max(0, pct)
    pct = min(100, pct)
    return str(pct)


def decode_x(s: str) -> str:
    """Take a unicode string of the form "b\xc3\xbct" and fix it"""
    s = re.sub("\x00", "", s, count=0, flags=re.I)
    s = re.sub("'", "\\'", s, count=0, flags=re.I)
    s = eval("'{}'".format(s))
    outlist = []
    for cp in s:
        num = ord(cp)
        if num < 255:
            outlist.append(struct.pack("B", num))
        elif num < 65535:
            outlist.append(struct.pack(">H", num))
        else:
            b = (num & 0xFF0000) >> 16
            H = num & 0xFFFF
            outlist.append(struct.pack(">bH", b, H))
    blist = b"".join(outlist)
    return blist.decode("utf-8")


def devaps(dev: str, dump: str = "") -> List[Dict[str, str]]:
    """Get a list of Access Points (as dicts) for a device"""
    if not dump:
        log.debug("Getting AP list from 'iw' dev %s" % dev)
        dump = docmd("iw dev %s scan" % dev)

    aps = []
    for blk in re.split("\nBSS ", dump[4:]):
        try:
            ap = blk2dict(blk)
            ap["power"] = dbm2pct(float(ap["signal"].split()[0]))

            if ap["SSID"] and ap["SSID"].count(r"\x00") == 0:
                ap["SSID"] = decode_x(ap["SSID"])
                aps.append(ap)
        except KeyError:
            pass

    return aps


def apgen(dev: str, q: Queue, dump: str = "") -> None:
    for ap in devaps(dev, dump):
        pt = {}
        pt["ssid"] = ap["SSID"]
        pt["strength"] = ap["power"]
        if "WPA" in ap or "RSN" in ap:
            pt["security"] = "encrypted"
        else:
            pt["security"] = "unencrypted"

        q.put(pt)

    q.put("DONE")


def dedup_aplist(aplist: List[Dict[str, str]]) -> List[Dict[str, str]]:
    apdict = {x["ssid"]: x for x in aplist}

    return [apdict[x] for x in apdict]


def candidates(device: str = None) -> List[Dict[str, str]]:
    """Return a list of reachable Access Point SSIDs, sorted by power"""

    if device:
        dev_list = [device]
    else:
        dev_list = devlist()

    jobs = []
    q: Queue = Queue()
    for dev in dev_list:
        p = Process(target=apgen, args=(dev, q))
        p.start()
        jobs.append(p)

    clist = []
    donecount = 0
    while donecount < len(jobs):
        pt = q.get()
        if pt == "DONE":
            donecount += 1
        else:
            clist.append(pt)

    for p in jobs:
        p.join()

    clist = dedup_aplist(clist)

    clist = sorted(clist, key=lambda x: -float(x["strength"]))

    return clist


def ap_conn_count() -> int:
    count = 0
    for dev in devlist():
        log.debug("Getting iw station dump")
        out = docmd("iw dev %s station dump" % dev)
        count += len([x for x in out.split("\n") if "Station" in x])

    return count


if __name__ == "__main__":
    print(candidates())

    print(ap_conn_count())

    import sys

    for ap in devaps(sys.argv[1], sys.stdin.read()):
        print(ap["SSID"], ap["RSN"], ap["power"], "RSN" in ap)



from typing import NamedTuple

import pytest

from comitup.iwscan import decode_x


class Case(NamedTuple):
    instr: str
    outstr: str


@pytest.mark.parametrize(
    "case",
    [
        Case("but", "but"),
        Case("b\xc3\xbct", "büt"),
        Case("b\x00\x00t", "bt"),
    ],
)
def test_decode_x(case):
    assert decode_x(case.instr) == case.outstr



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import logging
import sys
import time
import urllib
from logging.handlers import TimedRotatingFileHandler
from multiprocessing import Process

from cachetools import TTLCache, cached
from flask import (
    Flask,
    abort,
    jsonify,
    redirect,
    render_template,
    request,
    send_from_directory,
)

sys.path.append(".")
sys.path.append("..")

from comitup import client as ciu  # noqa

ciu_client = None
LOG_PATH = "/var/log/comitup-web.log"
TEMPLATE_PATH = "/usr/share/comitup/web/templates"

ttl_cache: TTLCache = TTLCache(maxsize=10, ttl=5)


def deflog():
    log = logging.getLogger("comitup_web")
    log.setLevel(logging.INFO)
    handler = TimedRotatingFileHandler(
        LOG_PATH,
        encoding="utf=8",
        when="W0",
        backupCount=8,
    )
    fmtr = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(fmtr)
    log.addHandler(handler)

    return log


def do_connect(ssid, password, log):
    time.sleep(1)
    log.debug("Calling client connect")
    ciu_client.service = None
    ciu_client.ciu_connect(ssid, password)


@cached(cache=ttl_cache)
def cached_points():
    return ciu_client.ciu_points()


def create_app(log):
    app = Flask(__name__, template_folder=TEMPLATE_PATH)

    @app.after_request
    def add_header(response):
        response.cache_control.max_age = 0
        return response

    @app.route("/")
    def index():
        points = cached_points()
        for point in points:
            point["ssid_encoded"] = urllib.parse.quote(point["ssid"])
        log.info("index.html - {} points".format(len(points)))
        return render_template(
            "index.html", points=points, can_blink=ciu.can_blink()
        )

    @app.route("/confirm")
    def confirm():
        ssid = request.args.get("ssid", "")
        ssid_encoded = urllib.parse.quote(ssid.encode())
        encrypted = request.args.get("encrypted", "unencrypted")

        mode = ciu_client.ciu_info()["imode"]

        log.info("confirm.html - ssid {0}, mode {1}".format(ssid, mode))

        return render_template(
            "confirm.html",
            ssid=ssid,
            encrypted=encrypted,
            ssid_encoded=ssid_encoded,
            mode=mode,
            can_blink=ciu.can_blink(),
        )

    @app.route("/connect", methods=["POST"])
    def connect():
        ssid = urllib.parse.unquote(request.form["ssid"])
        password = request.form["password"].encode()

        cached_points()

        p = Process(target=do_connect, args=(ssid, password, log))
        p.start()

        log.info("connect.html - ssid {0}".format(ssid))
        return render_template(
            "connect.html",
            ssid=ssid,
            password=password,
        )

    @app.route("/blink")
    def blink():
        log.info("blinking")
        ciu.blink()

        resp = jsonify(success=True)
        return resp

    @app.route("/img/favicon.ico")
    def favicon():
        log.info("Returning 404 for favicon request")
        abort(404)

    @app.route("/img/<path:path>")
    def send_image(path):
        return send_from_directory(TEMPLATE_PATH + "/images", path)

    @app.route("/js/<path:path>")
    def send_js(path):
        return send_from_directory(TEMPLATE_PATH + "/js", path)

    @app.route("/css/<path:path>")
    def send_css(path):
        return send_from_directory(TEMPLATE_PATH + "/css", path)

    @app.route("/<path:path>")
    def catch_all(path):
        log.info("Redirecting {0}".format(path))
        return redirect("http://10.41.0.1/", code=302)

    @app.errorhandler(500)
    def internal_error(error):
        log.error("Internal Error detected")
        sys.exit(1)

    return app


def main():
    log = deflog()
    log.info("Starting comitup-web")

    global ciu_client
    ciu_client = ciu.CiuClient()

    ciu_client.ciu_state()
    ciu_client.ciu_points()

    app = create_app(log)
    app.run(host="0.0.0.0", port=80, debug=False, threaded=True)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE


import logging
import os
import re
import subprocess
import textwrap
from collections import namedtuple
from typing import List, Optional, Tuple

log = logging.getLogger("comitup")


class DevInfo(object):
    def __init__(self):
        self.dev_list: List[Tuple[str, str]] = []
        for dev in os.listdir("/sys/class/net"):
            try:
                path = "/sys/class/net/{}/phy80211/name".format(dev)
                with open(path, "r") as fp:
                    phy = fp.read().strip()
                self.dev_list.append((dev, phy))
            except (NotADirectoryError, FileNotFoundError):
                pass

    def get_devs(self) -> List[str]:
        return sorted([x[0] for x in self.dev_list])

    def get_phy(self, dev: str) -> str:
        return [x[1] for x in self.dev_list if x[0] == dev][0]


dev_info: DevInfo = DevInfo()


def device_present() -> Optional[str]:
    if dev_info.get_devs():
        return None
    else:
        # Fail without comment
        return ""


def device_supports_ap() -> Optional[str]:
    dev: str = dev_info.get_devs()[0]
    phy: str = dev_info.get_phy(dev)

    try:
        cmd: str = "iw phy {} info".format(phy)
        deviceinfo: str = subprocess.check_output(cmd.split()).decode()
    except subprocess.CalledProcessError:
        return ""

    if "* AP\n" not in deviceinfo:
        return phy

    return None


def device_nm_managed() -> Optional[str]:
    try:
        cmd: str = "nmcli device show"
        try:
            devsinfo: str = subprocess.check_output(
                cmd.split(), re.MULTILINE
            ).decode()
        except UnicodeDecodeError:
            # shouldn't happen, but it does. Move on
            return None

        for dev in dev_info.get_devs():
            if dev not in devsinfo:
                # Fail without comment
                return ""
    except subprocess.CalledProcessError:
        pass

    return None


testspec = namedtuple("testspec", ["testfn", "title", "description"])


testspecs = [
    testspec(
        device_present,
        "comitup-no-wifi - No wifi devices found",
        textwrap.dedent(
            """
            Comitup is a wifi device manager. 'sudo iw list' indicates that
            there are no devices to manage.
        """
        ),
    ),
    testspec(
        device_supports_ap,
        "comitup-no-ap - The Main wifi device doesn't support AP mode",
        textwrap.dedent(
            """
            Comitup uses the first wifi device to implement the comitup-<nnn>
            Access Point. For this to work, the device must include "AP" in
            list of "Supported interface modes" returned by "iw list".
        """
        ),
    ),
    testspec(
        device_nm_managed,
        "comitup-no-nm - Wifi device is not managed by NetworkManager",
        textwrap.dedent(
            """
            Comitup uses NetworkManager to manage the wifi devices, but the
            required devices are not listed. This usually means that the
            devices are listed in /etc/network/interfaces, and are therefore
            being managed elsewhere. Remove the references to wifi devices
            from that file.
        """
        ),
    ),
]


def run_checks(logit=True, printit=True, verbose=True) -> bool:
    for testspec in testspecs:
        testresult = testspec.testfn()
        if testresult is not None:
            if logit:
                log.error(testspec.title)
                if testresult:
                    log.error("    " + testresult)

            if printit:
                print(testspec.title)
                if testresult:
                    print("    " + testresult)
                if verbose:
                    print(testspec.description)
            return True

    return False


if __name__ == "__main__":
    run_checks(logit=False)



import logging
import os
import signal
import time
from functools import wraps

enabled = False
try:
    import RPi.GPIO as GPIO

    enabled = True
except ModuleNotFoundError:
    pass

from comitup import blink  # noqa
from comitup import config  # noqa
from comitup import nm  # noqa

GPIO_INPUT = 21

log = logging.getLogger("comitup")


def checkenabled(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        if not enabled:
            return

        (conf, _) = config.load_data()
        if not conf.getboolean("enable_nuke"):
            return

        return fn(*args, **kwargs)

    return wrapper


@checkenabled
def nuke():
    for ssid in nm.get_all_wifi_connection_ssids():
        nm.del_connection_by_ssid(ssid)

    os.kill(os.getpid(), signal.SIGTERM)


def gpio_callback(dummy):
    log.debug("Nuke start event detected")
    process_low_event()


def process_low_event():
    total_time = 3
    check_interval = 0.01

    for _ in range(int(total_time / check_interval)):
        time.sleep(check_interval)
        if GPIO.input(GPIO_INPUT):
            return

    log.warning("Nuke function invoked")

    blink.blink(3)

    nuke()


@checkenabled
def init_nuke():
    GPIO.setmode(GPIO.BCM)

    GPIO.setup(GPIO_INPUT, GPIO.IN, pull_up_down=GPIO.PUD_UP)

    GPIO.add_event_detect(GPIO_INPUT, GPIO.FALLING, gpio_callback, 10)

    # So maybe the pin is already shorted?
    process_low_event()


@checkenabled
def cleanup_nuke():
    GPIO.remove_event_detect(GPIO_INPUT)
    GPIO.setup(GPIO_INPUT, GPIO.IN, pull_up_down=GPIO.PUD_OFF)



from typing import Any, Dict, List

import dbus

NM_DEVICE_STATE_ACTIVATED: int
NM_DEVICE_STATE_FAILED: int
NM_DEVICE_STATE_DISCONNECTED: int
NM_DEVICE_STATE_REASON_NO_SECRETS: int

class ObjectVanished(Exception): ...
class NMDbusInterfaceType(type): ...

class NMDbusInterface(object):
    object_path: dbus.ObjectPath

class Device(NMDbusInterface):
    DeviceType: int
    Interface: str
    Ip4Address: str
    ActiveConnection: Connection
    @staticmethod
    def SpecificDevice() -> Device: ...
    @staticmethod
    def Disconnect() -> None: ...
    @staticmethod
    def GetAllAccessPoints() -> List[AccessPoint]: ...
    class Ip4Config:
        Addresses: List[List[str]]
    class Ip6Config:
        Addresses: List[List[str]]

class Wireless(Device): ...

class NetworkManager(NMDbusInterface):
    State: int
    @staticmethod
    def GetDevices() -> List[Device]: ...
    @staticmethod
    def ActivateConnection(
        connection: Connection, dev: Device, path: str
    ) -> Connection: ...
    @staticmethod
    def DeactivateConnection(connection: Connection) -> Connection: ...

class Settings(NMDbusInterface):
    @staticmethod
    def ListConnections() -> List[Connection]: ...
    @staticmethod
    def ReloadConnections() -> None: ...
    @staticmethod
    def AddConnection(settings: Dict[str, Any]) -> None: ...

class Connection(NMDbusInterface):
    uuid: str
    Connection: Any
    @staticmethod
    def Delete() -> None: ...
    @staticmethod
    def GetSettings() -> Dict[str, Any]: ...

class AccessPoint(NMDbusInterface):
    Strength: int
    Ssid: str
    Flags: int
    WpaFlags: int
    RsnFlags: int



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
import os
import textwrap
from collections import namedtuple

import pytest

from comitup import config


@pytest.fixture()
def conf_fxt(tmpdir):
    path = os.path.join(tmpdir.__str__(), "conf.conf")

    with open(path, "w") as fp:
        fp.write(
            textwrap.dedent(
                """
                #
                # comment
                #
                tag1=val1
                tag2: val2
                tag3 = val3
                # tag4 = val4
                """
            )
        )

    return config.Config(path)


@pytest.mark.parametrize("idx", ("1", "2", "3"))
def test_conf_vals(idx, conf_fxt):
    assert eval('conf_fxt.tag{0} == "val{0}"'.format(idx))


def test_conf_miss(conf_fxt):
    with pytest.raises(AttributeError):
        conf_fxt.tag4


Case = namedtuple("Case", ["data", "result"])
BoolFixt = namedtuple("BoolFixt", ["config", "result"])


@pytest.fixture(
    params=[
        Case("0", False),
        Case("1", True),
        Case("yes", True),
        Case("Yes", True),
        Case("no", False),
        Case("y", True),
        Case("n", False),
        Case("true", True),
        Case("True", True),
        Case("TRUE", True),
        Case("false", False),
        Case("on", True),
        Case("off", False),
    ]
)
def bool_fixt(tmpdir, request):
    path = os.path.join(tmpdir.__str__(), "conf.conf")

    with open(path, "w") as fp:
        fp.write("tag = {}".format(request.param.data))

    return BoolFixt(config.Config(path), request.param.result)


def test_conf_bool(bool_fixt):
    assert bool_fixt.config.getboolean("tag") == bool_fixt.result



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import pytest
from mock import call, patch

from comitup import webmgr


@pytest.fixture()
def websvc_fxt(request):
    web_svc = webmgr.web_service

    def fin():
        webmgr.web_service = web_svc

    request.addfinalizer(fin)


@pytest.mark.parametrize(
    "state, action, fn_fact, arg_fact",
    (
        (
            "HOTSPOT",
            "start",
            lambda: webmgr.stop_service,
            lambda: webmgr.web_service,
        ),  # noqa
        (
            "HOTSPOT",
            "pass",
            lambda: webmgr.start_service,
            lambda: webmgr.COMITUP_SERVICE,
        ),  # noqa
        (
            "CONNECTING",
            "start",
            lambda: webmgr.stop_service,
            lambda: webmgr.COMITUP_SERVICE,
        ),  # noqa
        (
            "CONNECTED",
            "start",
            lambda: webmgr.start_service,
            lambda: webmgr.web_service,
        ),  # noqa
    ),
)  # noqa
@pytest.mark.parametrize("svc", ("", "foo"))
@patch("comitup.webmgr.start_service")
@patch("comitup.webmgr.stop_service")
def test_webmgr_callback(
    stop_svc, start_svc, svc, state, action, fn_fact, arg_fact, websvc_fxt
):
    webmgr.web_service = svc
    webmgr.state_callback(state, action)

    if arg_fact():
        assert fn_fact().called
        assert fn_fact().called_with(call(arg_fact()))
    else:
        assert not fn_fact().called

    if svc:
        assert fn_fact().called


others = [
    (x, y)
    for x in ("HOTSPOT", "CONNECTING", "CONNECTED")
    for y in ("fail", "timeout")
]  # noqa


@pytest.mark.parametrize(
    "state, action",
    [
        ("CONNECTING", "pass"),
        ("CONNECTED", "pass"),
    ]
    + others,
)
@patch("comitup.webmgr.start_service")
@patch("comitup.webmgr.stop_service")
def test_webmgr_no_callback(stop_svc, start_svc, state, action, websvc_fxt):
    webmgr.web_service = "foo"
    webmgr.state_callback(state, action)

    assert not stop_svc.called
    assert not start_svc.called



# Copyright (c) 2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import dbus

bus = dbus.SystemBus()
systemd_service = bus.get_object(
    "org.freedesktop.systemd1",
    "/org/freedesktop/systemd1",
)

sd_start_unit = systemd_service.get_dbus_method(
    "StartUnit",
    "org.freedesktop.systemd1.Manager",
)

sd_stop_unit = systemd_service.get_dbus_method(
    "StopUnit",
    "org.freedesktop.systemd1.Manager",
)

sd_unit_state = systemd_service.get_dbus_method(
    "GetUnitFileState",
    "org.freedesktop.systemd1.Manager",
)

sd_unit_jobs = systemd_service.get_dbus_method(
    "GetUnitProcesses",
    "org.freedesktop.systemd1.Manager",
)

if __name__ == "__main__":
    print(sd_unit_state("NetworkManager.service"))



# Copyright (c) 2018-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import pathlib
import urllib

import pytest
from mock import Mock

from web import comitupweb

ssid_list = [
    "simple",
    "simple+",
    "simplé",
    "simple 2",
    "simplE",
]


@pytest.fixture(params=ssid_list)
def ssid(request):
    return request.param


@pytest.fixture
def app(monkeypatch):
    templatedir = pathlib.Path(__file__).parent.parent / "web/templates"
    monkeypatch.setattr("web.comitupweb.TEMPLATE_PATH", str(templatedir))

    app = comitupweb.create_app(Mock())
    app.debug = True
    app.testing = True
    return app


def test_webapp_null(app, ssid):
    assert "simpl" in ssid


def test_webapp_index(app, ssid, monkeypatch):
    comitupweb.ttl_cache.clear()

    point = {
        "ssid": ssid,
    }
    client_mock = Mock()
    client_mock.ciu_points.return_value = [point]
    monkeypatch.setattr("web.comitupweb.ciu_client", client_mock)
    # comitupweb.ciu_client = client_mock

    response = app.test_client().get("/")
    index_text = response.get_data().decode()

    assert ssid + "</button>" in index_text
    assert "ssid=" + urllib.parse.quote(ssid) in index_text


def test_webapp_connect(app, ssid, monkeypatch):
    monkeypatch.setattr("web.comitupweb.Process", Mock())

    data = {
        "ssid": urllib.parse.quote(ssid),
        "password": urllib.parse.quote("password"),
    }
    response = app.test_client().post("connect", data=data)
    index_text = response.get_data().decode()

    assert "to " + ssid in index_text


@pytest.mark.parametrize("canblink", [True, False])
@pytest.mark.parametrize("path", ["/confirm"])
def test_webapp_blink_confirm(app, canblink, monkeypatch, path):
    monkeypatch.setattr("web.comitupweb.ciu.can_blink", lambda: canblink)
    monkeypatch.setattr("web.comitupweb.ciu.blink", Mock())
    monkeypatch.setattr("comitup.blink.set_trigger", Mock())
    monkeypatch.setattr("web.comitupweb.ciu_client", Mock())
    monkeypatch.setattr(
        "web.comitupweb.ciu_client.ciu_info",
        Mock(return_value={"imode": "single"}),
    )
    monkeypatch.setattr(
        "web.comitupweb.ciu_client.ciu_points", Mock(return_value=[])
    )

    response = app.test_client().get(path)
    blinktext = response.get_data().decode()

    assert canblink == ("Locate" in blinktext)


def test_webapp_blink(app, monkeypatch):
    monkeypatch.setattr("comitup.blink.set_trigger", Mock())
    monkeypatch.setattr("web.comitupweb.ciu.blink", Mock())
    response = app.test_client().get("/blink")

    assert response.status_code == 200



#!/usr/bin/python3
# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#


import hashlib
import logging
from functools import wraps
from typing import Callable, List, Optional

import dbus
import NetworkManager
from gi.repository.GLib import timeout_add

from comitup import iwscan, mdns, modemgr, nm, nmmon, routemgr, wpa

log = logging.getLogger("comitup")

# constants
DEFAULT_TIMEOUT = 180  # default time for state timeout functions
FAIL_TIMEOUT = 360  # hotspot timeout after a connecting fail

# Global state information
dns_names: List[str] = []
com_state: Optional[str] = None
conn_list: List[str] = []
connection: str = ""
state_id: int = 0
startup: bool = False

state_callbacks: List[Callable[[str, str], None]] = []

hotspot_name: str = ""


def state_callback(fn: Callable[[int], None]) -> Callable[[int], None]:
    @wraps(fn)
    def wrapper(reason: int) -> None:
        state: str
        action: str
        state, action = fn.__name__.split("_")

        log.debug("State call - {}-{}".format(state, action))

        state = state.upper()

        fn(reason)

        call_callbacks(state, action)

    return wrapper


def call_callbacks(state: str, action: str) -> None:
    if state not in ["HOTSPOT", "CONNECTING", "CONNECTED"]:
        log.debug("Illegal state {}".format(state))

    if action not in ["start", "pass", "fail", "timeout"]:
        log.debug("Illegal action {}".format(action))

    log.debug("Calling callbacks for {}:{}".format(state, action))
    for callback in state_callbacks:
        callback(state, action)
    log.debug("Callbacks complete")


def timeout(fn: Callable[[int], None]) -> Callable[[int, int], bool]:
    @wraps(fn)
    def wrapper(id: int, reason: int) -> bool:
        if id == state_id:
            fn(reason)
            return True
        else:
            return False

    return wrapper


def dns_to_conn(host: str) -> str:
    if ".local" in host:
        return host[: -len(".local")]
    else:
        return host


#
# Hotspot state
#


def fake_hs_pass(sid: int) -> bool:
    hotspot_pass(sid, 0)
    return False


@state_callback
def hotspot_start(dummy: int) -> None:
    global conn_list
    log.info("Activating hotspot")

    if startup:
        mdns.clear_entries()
        mdns.add_hosts(dns_names)

    hs_ssid: str = dns_to_conn(dns_names[0])

    if startup and modemgr.get_mode() == modemgr.SINGLE_MODE:
        log.debug("Passing on hotspot connection for now")
        timeout_add(100, fake_hs_pass, state_id)
    elif hs_ssid != nm.get_active_ssid(modemgr.get_state_device("HOTSPOT")):
        conn_list = []

        log.debug("Activating connection {}".format(hs_ssid))
        activate_connection(hs_ssid, "HOTSPOT")
    else:
        log.debug("Didn't need to reactivate - already running")
        # the connect callback won't happen - let's 'pass' manually
        timeout_add(100, fake_hs_pass, state_id)


@timeout
@state_callback
def hotspot_pass(reason: int) -> None:
    global startup

    dev: NetworkManager.Device = modemgr.get_state_device("CONNECTED")
    conn_list = candidate_connections(dev)
    active_ssid = nm.get_active_ssid(modemgr.get_state_device("CONNECTED"))
    if startup or active_ssid in conn_list:
        set_state("CONNECTING", conn_list)
        startup = False


@timeout
@state_callback
def hotspot_fail(reason: int) -> None:
    log.warning("Hotspot mode failure")
    set_state("HOTSPOT", force=True)


@timeout
def hotspot_timeout(dummy: int) -> None:
    if iwscan.ap_conn_count() == 0 or modemgr.get_mode() != "single":
        log.debug("Periodic connection attempt")

        dev = modemgr.get_state_device("CONNECTED")
        conn_list: List[str] = candidate_connections(dev)
        if conn_list:
            set_state("CONNECTING", conn_list)
        else:
            log.info("No candidates - skipping CONNECTING scan")
    else:
        log.info("AP active - skipping CONNECTING scan")

    wpa.check_wpa(modemgr.get_ap_device().Interface)


#
# Connecting state
#


def fake_cg_pass(sid: int) -> bool:
    connecting_pass(sid, 0)
    return False


@state_callback
def connecting_start(dummy: int) -> None:
    global conn_list, connection

    dev = modemgr.get_state_device("CONNECTED")
    full_conn_list = candidate_connections(dev)
    active_ssid = nm.get_active_ssid(modemgr.get_state_device("CONNECTED"))
    if active_ssid in full_conn_list:
        log.debug("Didn't need to connect - already connected")
        connection = active_ssid
        # the connect callback won't happen - let's 'pass' manually
        timeout_add(100, fake_cg_pass, state_id)
    else:
        if conn_list:
            log.debug("states: Calling nm.disconnect()")
            nm.disconnect(modemgr.get_state_device("CONNECTING"))

            conn = conn_list.pop(0)
            log.info("Attempting connection to %s" % conn)
            try:
                activate_connection(conn, "CONNECTING")
            except dbus.exceptions.DBusException:
                msg = "DBUS failure connecting to {} - skipping".format(conn)
                log.warning(msg)
                connecting_fail(state_id, 0)
        else:
            set_state("HOTSPOT")


@timeout
@state_callback
def connecting_pass(reason: int) -> None:
    log.debug("Connection successful")
    set_state("CONNECTED")


@timeout
@state_callback
def connecting_fail(reason: int) -> None:
    global conn_list

    log.debug("Connection failed - reason {}".format(reason))

    badreasons: List[int] = [
        # NetworkManager.NM_DEVICE_STATE_REASON_NO_SECRETS,  # Disable delete
    ]
    if reason in badreasons:
        log.error("Connection {} config failure - DELETING".format(connection))
        nm.del_connection_by_ssid(connection)

    if conn_list:
        set_state("CONNECTING", force=True)
    else:
        set_state("HOTSPOT", timeout=FAIL_TIMEOUT)


@timeout
def connecting_timeout(dummy: int) -> None:
    connecting_fail(state_id, 0)


#
# Connect state
#


def fake_cn_pass(sid: int) -> bool:
    connected_pass(sid, 0)
    return False


@state_callback
def connected_start(dummy: int) -> None:
    global conn_list

    conn_list = []

    # Connected mode always passes
    timeout_add(100, fake_cn_pass, state_id)


@timeout
@state_callback
def connected_pass(reason: int) -> None:
    pass


@timeout
@state_callback
def connected_fail(reason: int) -> None:
    global startup
    log.warning("Connection lost")

    active_ssid: Optional[str]
    active_ssid = nm.get_active_ssid(modemgr.get_state_device("HOTSPOT"))
    if modemgr.get_mode() == modemgr.MULTI_MODE and not active_ssid:
        log.warning("Hotspot lost while CONNECTED")
        set_state("HOTSPOT")
    else:
        startup = True
        set_state("HOTSPOT")


@timeout
def connected_timeout(dummy: int) -> None:
    active_ssid: Optional[str]
    active_ssid = nm.get_active_ssid(modemgr.get_state_device("CONNECTED"))
    log.debug(
        "connected_timeout comparing {} to {}".format(connection, active_ssid)
    )
    if connection != active_ssid:
        log.warning("Connection lost on timeout")
        dev = modemgr.get_state_device("CONNECTED")
        set_state("CONNECTING", candidate_connections(dev))

    if modemgr.get_mode() == modemgr.MULTI_MODE:
        wpa.check_wpa(modemgr.get_ap_device().Interface)

        active_ssid = nm.get_active_ssid(modemgr.get_state_device("HOTSPOT"))
        if not active_ssid:
            log.warning("Hotspot lost on timeout")
            set_state("HOTSPOT")

        defroute_devname: Optional[str] = routemgr.defroute_dev()
        ap_dev: NetworkManager.Device = modemgr.get_ap_device()
        link_dev: NetworkManager.Device = modemgr.get_link_device()
        if defroute_devname == nm.device_name(ap_dev):
            # default route is bad. Disconnect link and count on state
            # processing to restore
            log.error("AP is holding default route while CONNECTED, kicking")
            nm.disconnect(link_dev)

    # Leave this out for now. Disabling dhcpcd may have fixed the instability
    # problem.
    # mdns.check_mdns(dns_names)


#
# State Management
#


class state_matrix(object):
    """Map e.g. state_matrix('HOTSPOT').pass_fn to the function hotspot_pass"""

    def __init__(self, state: str):
        self.state: str = state.lower()

    def __getattr__(self, attr):
        try:
            fname: str = self.state + "_" + attr[:-3]
            return globals()[fname]
        except KeyError:
            print(attr)
            raise AttributeError


def set_state(
    state: str,
    connections: List[str] = [],
    timeout: int = DEFAULT_TIMEOUT,
    force: bool = False,
) -> None:
    timeout_add(0, set_state_to, state, connections, timeout, force, state_id)


def set_state_to(
    state: str,
    connections: List[str],
    timeout: int,
    force: bool,
    curr_state_id: int,
):
    global com_state, conn_list, state_id

    if state == com_state and not force:
        return False

    if curr_state_id < state_id:
        return False

    log.info("Setting state to %s" % state)

    state_info: state_matrix = state_matrix(state)

    nmmon.init_nmmon()

    state_id += 1

    nmmon.enable(
        modemgr.get_state_device(state),
        state_info.pass_fn,
        state_info.fail_fn,
        state_id,
    )

    if state in ["CONNECTED", "HOTSPOT"]:
        nmmon.enhance_fail_states()

    if connections:
        conn_list = connections

    com_state = state
    timeout_add(timeout * 1000, state_info.timeout_fn, state_id, 0)
    state_info.start_fn(0)

    return False


def activate_connection(name: str, state: str) -> None:
    global connection
    connection = name
    log.debug("Connecting to %s" % connection)

    path = "/"

    nm.activate_connection_by_ssid(
        connection, modemgr.get_state_device(state), path=path
    )


def candidate_connections(device: NetworkManager.Device) -> List[str]:
    return nm.get_candidate_connections(device)


def set_hosts(*args):
    global dns_names
    dns_names = args


def hash_conf() -> str:
    m = hashlib.sha256()
    with open("/etc/comitup.conf", "rb") as fp:
        m.update(fp.read())

    return m.hexdigest()[-4:]


def assure_hotspot(
    ssid: str, device: NetworkManager.Device, password: str
) -> None:
    nm.del_connection_by_ssid(ssid)
    if not nm.get_connection_by_ssid(ssid):
        nm.make_hotspot(ssid, device, password)


def init_states(
    hosts: List[str],
    callbacks: List[Callable[[str, str], None]],
    hotspot_pw: str,
) -> None:
    global hotspot_name, conn_list, connection, startup

    nmmon.init_nmmon()
    set_hosts(*hosts)

    for callback in callbacks:
        add_state_callback(callback)

    hotspot_name = dns_to_conn(hosts[0])
    assure_hotspot(hotspot_name, modemgr.get_ap_device(), hotspot_pw)

    startup = True
    set_state("HOTSPOT")


def add_state_callback(callback: Callable[[str, str], None]) -> None:
    global state_callbacks

    state_callbacks.append(callback)



import pytest

# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
from mock import Mock, patch

from comitup import nm


@pytest.fixture()
def no_device_fxt(monkeypatch):
    monkeypatch.setattr(
        "comitup.nm.nm.NetworkManager.GetDevices", Mock(return_value=[])
    )
    return None


@pytest.fixture()
def device_no_conn_fxt(monkeypatch):
    device = Mock()
    device.ActiveConnection = None
    monkeypatch.setattr(
        "comitup.nm.nm.NetworkManager.GetDevices", Mock(return_value=[device])
    )
    return None


@pytest.fixture()
def device_fxt(monkeypatch):
    device = Mock()
    device.DeviceType = 2
    device.ActiveConnection.Connection.GetSettings.return_value = {
        "802-11-wireless": {
            "ssid": "myssid",
        }
    }
    device.Ip4Address = "1.2.3.4"

    point = Mock()
    point.Ssid = "myssid"
    point.Strength = 100
    getAllAccessPoints = Mock()
    getAllAccessPoints.GetAllAccessPoints.return_value = [point]
    device.SpecificDevice.return_value = getAllAccessPoints

    monkeypatch.setattr(
        "comitup.nm.nm.NetworkManager.GetDevices", Mock(return_value=[device])
    )

    return None


@pytest.fixture()
def no_connections_fxt(monkeypatch):
    monkeypatch.setattr(
        "comitup.nm.nm.Settings.ListConnections", Mock(return_value=[])
    )
    return None


@pytest.fixture()
def connections_fxt(monkeypatch):
    connection = Mock()
    connection.GetSettings.return_value = {
        "802-11-wireless": {
            "ssid": "myssid",
        }
    }

    monkeypatch.setattr(
        "comitup.nm.nm.Settings.ListConnections",
        Mock(return_value=[connection]),
    )

    return connection


def test_get_active_ssid(device_fxt):
    assert nm.get_active_ssid(nm.get_wifi_device()) == "myssid"


def test_get_active_ip(device_fxt):
    assert nm.get_active_ip(nm.get_wifi_device()) == "1.2.3.4"


def test_no_conn(no_connections_fxt):
    assert nm.get_connection_by_ssid("ssid") is None


def test_get_connection_by_ssid(connections_fxt):
    assert nm.get_connection_by_ssid("myssid")
    assert not nm.get_connection_by_ssid("bogusssid")


def test_del_connection_by_ssid(connections_fxt):
    nm.del_connection_by_ssid("myssid")
    assert connections_fxt.Delete.called


@patch("comitup.nm.get_wifi_device")
def test_activate_connection_by_id(get_dev, monkeypatch, connections_fxt):
    activate = Mock()
    monkeypatch.setattr(
        "comitup.nm.nm.NetworkManager.ActivateConnection", activate
    )

    nm.activate_connection_by_ssid("myssid", nm.get_wifi_device())
    assert activate.called


def test_make_hotspot(monkeypatch):
    addconnection = Mock()
    monkeypatch.setattr("comitup.nm.nm.Settings.AddConnection", addconnection)

    nm.make_hotspot()

    assert addconnection.called


def test_make_connection_for(monkeypatch):

    addconnection = Mock()
    monkeypatch.setattr("comitup.nm.nm.Settings.AddConnection", addconnection)

    nm.make_connection_for("anssid", "password")

    assert addconnection.called


def test_do_listaccess(device_fxt):
    nm.do_listaccess(None)



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE
#
# Copyright 2016-2017 David Steele <steele@debian.org>
# This file is part of comitup
# Available under the terms of the GNU General Public License version 2
# or later
#

import json
import os
from functools import wraps


def persist_decorator(klass):
    """Add a save behavior to methods that update dict data"""
    for method in ["__setitem__", "__delitem__", "update", "setdefault"]:
        setattr(klass, method, klass._addsave(getattr(klass, method)))

    return klass


@persist_decorator
class persist(dict):
    """A JSON-file backed persistent dictionary"""

    def __init__(self, path, *args, **kwargs):
        """Initialize with backing file path, and optional dict defaults"""

        super(persist, self).__init__(*args, **kwargs)

        self._path = path

        if os.path.exists(self._path):
            self._load()

        self._save()

    def _save(self):
        with open(self._path, "w") as fp:
            json.dump(self, fp, indent=2)

    def _load(self):
        with open(self._path, "r") as fp:
            dct = json.load(fp)

        super().update(dct)

    def _addsave(fn):
        """Decorator to add save behavior to methods"""

        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            retval = fn(self, *args, **kwargs)
            self._save()
            return retval

        return wrapper

    def __setattr__(self, name, value):
        if name in self.__dict__ or name.startswith("_"):
            self.__dict__[name] = value
        else:
            self.__setitem__(name, value)

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            return self.__getitem__(name)



import textwrap

import pytest
from mock import Mock

from comitup import routemgr

routetxt = textwrap.dedent(
    """
        default via 192.168.200.1 dev ethn proto dhcp src 192.168.200.27
        default via 192.168.200.1 dev wlan1 proto dhcp src 192.168.200.177
        10.41.0.0/24 dev wlan0 proto kernel scope link src 10.41.0.1
        169.254.0.0/16 dev wlan0 scope link src 169.254.30.217 metric 303
        192.168.200.0/24 dev eth0 proto dhcp scope link src 192.168.200.27
        192.168.200.0/24 dev wlan1 proto dhcp scope link src 192.168.200.177
    """
).strip()


@pytest.fixture
def defroute(monkeypatch):
    cp = Mock()
    cp.stdout = routetxt
    runmethod = Mock(return_value=cp)
    monkeypatch.setattr("comitup.routemgr.subprocess.run", runmethod)


def test_defroute_fixture(defroute):
    cp = routemgr.subprocess.run(
        "ip route",
        stdout=routemgr.subprocess.PIPE,
        shell=True,
        encoding="utf-8",
    )
    assert "ethn" in cp.stdout


def test_default_dev(defroute):
    assert routemgr.defroute_dev() == "ethn"



# Copyright (c) 2017-2019 David Steele <dsteele@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later
# License-Filename: LICENSE

import pytest
from mock import Mock, call

from comitup import states


@pytest.fixture()
def state_globals(request):
    callbacks = states.state_callbacks

    def fin():
        states.state_callbacks = callbacks

    request.addfinalizer(fin)


@pytest.fixture()
def state_fxt(monkeypatch, state_globals):
    monkeypatches = (
        ("comitup.states.mdns.clear_entries", None),
        ("comitup.states.mdns.add_hosts", None),
        ("comitup.states.nm.activate_connection_by_ssid", None),
        ("comitup.states.nm.get_candidate_connections", ["c1", "c2"]),
        ("comitup.states.nm.get_active_ip", "1.2.3.4"),
        ("comitup.states.nm.get_active_ssid", None),
        ("comitup.states.nm.deactivate_connection", None),
        ("comitup.states.nmmon.init_nmmon", None),
        ("comitup.states.nmmon.set_device_callbacks", None),
        ("comitup.states.timeout_add", None),
        ("comitup.states.time.sleep", None),
    )

    for path, return_val in monkeypatches:
        if return_val:
            monkeypatch.setattr(path, Mock(return_value=return_val))
        else:
            monkeypatch.setattr(path, Mock())

    monkeypatch.setattr(
        "comitup.states.iwscan.ap_conn_count", Mock(return_value=0)
    )

    monkeypatch.setattr("comitup.states.modemgr.CONF_PATH", "/dev/null")

    states.set_hosts("hs", "hs-1111")


@pytest.fixture()
def points_fxt(monkeypatch):
    pt = {
        "nmpath": "/",
        "ssid": "ssid",
    }

    monkeypatch.setattr(
        "comitup.states.nm.get_points_ext", Mock(return_value=[pt])
    )

    return None


@pytest.mark.parametrize("offset, match", ((-1, False), (0, True), (1, False)))
def test_state_timeout_wrapper(offset, match):

    themock = Mock()

    @states.timeout
    def timeout_fn(dummy):
        themock()

    assert timeout_fn(states.state_id + offset, 0) == match
    assert themock.called == match


def test_state_timeout_activity():
    themock = Mock()

    @states.timeout
    def timeout_fn(dummy):
        themock()

    timeout_fn(states.state_id, 0)

    assert themock.called


def test_state_set_hosts():
    states.set_hosts("a", "b")
    assert states.dns_names == ("a", "b")


@pytest.mark.parametrize(
    "hostin, hostout", (("host", "host"), ("host.local", "host"))
)
def test_state_dns_to_conn(hostin, hostout):
    assert states.dns_to_conn(hostin) == hostout


def test_state_callback_decorator(state_globals):
    callback = Mock()

    @states.state_callback
    def foo_bar(dummy):
        pass

    states.add_state_callback(callback)

    foo_bar(0)

    assert callback.call_args == call("FOO", "bar")


def test_state_matrix():
    assert states.state_matrix("HOTSPOT").pass_fn == states.hotspot_pass


@pytest.mark.parametrize(
    "path, good",
    (
        ("states.state_matrix('HOTSPOT').bogus_fn", False),
        ("states.state_matrix('HOTSPOT').bogus", False),
        ("states.state_matrix('HOTSPOT').pass_fn", True),
    ),
)
def test_state_matrix_miss(path, good):
    if good:
        eval(path)
    else:
        with pytest.raises(AttributeError):
            eval(path)
